/*
 *   jack-ladspa-host
 *    
 *   Copyright (C) Robert Ham 2002 (node@users.sourceforge.net)
 *    
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <string.h>
#include <ctype.h>

#include <gtk/gtk.h>
#include <ladspa.h>

#include "gui.h"
#include "xmalloc.h"
#include "lock-free-fifo.h"
#include "callbacks.h"
#include "config.h"
#include "control-message.h"

#define TEXT_BOX_WIDTH        75
#define CONTROL_FIFO_SIZE     256
#define TEXT_BOX_CHARS        -1

GtkWidget * create_plugin_menu (plugin_collection_t * collection,
                                GCallback callback,
                                gpointer userdata) {
  GtkWidget * menu, * submenu, * menu_item, * submenu_item;
  plugin_t * plugin;
  unsigned long i;
  const char * last_name;
  char letter_str[2];
  char * str;
  char * filename;
  
  letter_str[1] = '\0';
  last_name = NULL;
  submenu = NULL;

  menu = gtk_menu_new ();
  
  for (i = 0; i < collection->plugin_count; i++) {
    plugin = collection->plugin_array[i];
    if (!last_name || strncasecmp (last_name, plugin->name, 1) != 0) { /* letter boundary */
      letter_str[0] = toupper (plugin->name[0]);
      menu_item = gtk_menu_item_new_with_label (letter_str);
      gtk_widget_show (menu_item);
      gtk_menu_shell_append (GTK_MENU_SHELL (menu), menu_item);
      
      submenu = gtk_menu_new ();
      gtk_widget_show (submenu);
      gtk_menu_item_set_submenu (GTK_MENU_ITEM (menu_item), submenu);
    }
    
    filename = strrchr (plugin->object_file->filename, '/');
    if (filename) {
      filename++;
    } else {
      filename = plugin->object_file->filename;
    }
    if (LADSPA_IS_HARD_RT_CAPABLE(plugin->descriptor->Properties)) {
      str = g_strdup_printf ("%s (%s, %s)", plugin->name, filename,
                             plugin->mono ? "mono" : "stereo");
    } else {
      str = g_strdup_printf ("%s (%s, %s, NOT RT)", plugin->name, filename,
                             plugin->mono ? "mono" : "stereo");
    }
    
    submenu_item = gtk_menu_item_new_with_label (str);

    g_free (str);

    /* use the glib object data stuff to carry the plugin */
    g_object_set_data (G_OBJECT (submenu_item), "jack-rack-plugin", plugin);
    g_object_set_data (G_OBJECT (submenu_item), "jack-rack-plugin-index", GINT_TO_POINTER(((int)i)));
    
    g_signal_connect (G_OBJECT (submenu_item), "activate",
                      callback, userdata);
    gtk_widget_show (submenu_item);
    gtk_menu_shell_append (GTK_MENU_SHELL (submenu), submenu_item);
    
    last_name = plugin->name;
  }
  
  return menu;
}

int create_gui (plugin_collection_t * collection,
                guiinfo_t * gui_info) {
  GtkWidget * widget1, * widget2, * widget3, * menu;

  /* main window */
  gui_info->main_window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title (GTK_WINDOW (gui_info->main_window), PACKAGE_NAME);
  gtk_window_set_default_size (GTK_WINDOW (gui_info->main_window), 620, 460);
  
  
  /* button/viewport box */
  widget3 = gtk_vbox_new (FALSE, 0);
  gtk_widget_show (widget3);
  gtk_container_add (GTK_CONTAINER (gui_info->main_window), widget3);
  
  
  /* top button box and buttons */
  widget1 = gtk_hbox_new (FALSE, 0);
  gtk_widget_show (widget1);
  gtk_box_pack_start (GTK_BOX (widget3), widget1, FALSE, FALSE, 0);
  
  widget2 = gtk_button_new_with_label ("Quit");
  gtk_widget_show (widget2);
  gtk_box_pack_end (GTK_BOX (widget1), widget2, FALSE, FALSE, 0);
  g_signal_connect (G_OBJECT (widget2), "clicked",
                      G_CALLBACK (quit_callback), gui_info);

/*  widget2 = gtk_vseparator_new ();
  gtk_widget_show (widget2);
  gtk_box_pack_end (GTK_BOX (widget1), widget2, FALSE, FALSE, 4);
  
  widget2 = gtk_button_new_with_label ("Save");
  gtk_widget_show (widget2);
  gtk_box_pack_end (GTK_BOX (widget1), widget2, FALSE, FALSE, 0);
  g_signal_connect (G_OBJECT (widget2), "clicked",
                      G_CALLBACK (save_callback), gui_info);

  widget2 = gtk_button_new_with_label ("Load");
  gtk_widget_show (widget2);
  gtk_box_pack_end (GTK_BOX (widget1), widget2, FALSE, FALSE, 0);
  g_signal_connect (G_OBJECT (widget2), "clicked",
                      G_CALLBACK (load_callback), gui_info);*/


  widget2 = gtk_button_new_with_label ("Add");
  gtk_widget_show (widget2);
  gtk_box_pack_start (GTK_BOX (widget1), widget2, FALSE, FALSE, 0);
  menu = create_plugin_menu (collection,
                             G_CALLBACK (add_callback), gui_info);
  
  g_signal_connect_swapped (G_OBJECT (widget2), "event",
	                    G_CALLBACK (plugin_button_callback),
	                    G_OBJECT (menu));  
  gui_info->add = widget2;                      


  /* viewport thingy */  
  widget1 = gtk_scrolled_window_new (NULL, NULL);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (widget1),
                                  GTK_POLICY_AUTOMATIC,
                                  GTK_POLICY_AUTOMATIC);
  gtk_box_pack_start (GTK_BOX (widget3), widget1, TRUE, TRUE, 0);
  gtk_widget_show (widget1);
                                  
  widget2 = gtk_viewport_new (NULL, NULL);
  gtk_container_add (GTK_CONTAINER (widget1), widget2);
  gtk_widget_show (widget2);


  /* the main vbox for the plugins */
  widget1 = gtk_vbox_new (FALSE, 0);
  gtk_container_add (GTK_CONTAINER (widget2), widget1);
  gtk_widget_show (widget1);
  gui_info->plugin_box = widget1;
  
  gtk_widget_show (gui_info->main_window);
  
  
  /* other stuff in the gui_info */
  gui_info->slots = NULL;
  gui_info->slots_end = NULL;
  gui_info->moving_slot = FALSE;
  gui_info->slot_count = 0;
  gui_info->shutdown = FALSE;
  
  gui_info->plugin_collection = collection;
  
  gtk_idle_add (idle_callback, gui_info);
  
  return 0;
}




/** create a control for float ports */
GtkWidget * create_float_control (const LADSPA_Descriptor * descriptor,
                                  unsigned long port_index) {
  LADSPA_Data lower;
  LADSPA_Data upper;
  GtkWidget * widget;

  if (LADSPA_IS_HINT_SAMPLE_RATE (descriptor->PortRangeHints[port_index].HintDescriptor)) {
    lower = descriptor->PortRangeHints[port_index].LowerBound * (LADSPA_Data) sample_rate;
    upper = descriptor->PortRangeHints[port_index].UpperBound * (LADSPA_Data) sample_rate;
  } else {
    lower = descriptor->PortRangeHints[port_index].LowerBound;
    upper = descriptor->PortRangeHints[port_index].UpperBound;
  }

  if (!LADSPA_IS_HINT_BOUNDED_BELOW (descriptor->PortRangeHints[port_index].HintDescriptor)) {
    lower = (LADSPA_Data) -100.0;
  }

  if (!LADSPA_IS_HINT_BOUNDED_ABOVE (descriptor->PortRangeHints[port_index].HintDescriptor)) {
    upper = (LADSPA_Data) 100.0;
  }
  
/*  printf ("%s: lower: %f, upper: %f, %s\n", __FUNCTION__, lower, upper,
          LADSPA_IS_HINT_SAMPLE_RATE (descriptor->PortRangeHints[port_index].HintDescriptor)
          ? "sample rate adjusted" : "normal");*/
  
  if (!(lower < upper)) {
    printf("%s: lower !< upper!\n", __FUNCTION__);
  }
  widget = gtk_hscale_new_with_range ((gdouble) lower, (gdouble) upper, (upper - lower) / 10.0);
  gtk_scale_set_draw_value (GTK_SCALE (widget), FALSE);
  gtk_scale_set_digits (GTK_SCALE (widget), 8);
  gtk_range_set_increments(GTK_RANGE(widget), (upper - lower) / 1000.0, (upper - lower) / 10.0);
  return widget;
}

/** create a control for int ports */
GtkWidget * create_int_control (const LADSPA_Descriptor * descriptor,
                                unsigned long port_index) {
  int lower = 0, upper = 0;
  GtkWidget * widget;
  
  if (LADSPA_IS_HINT_SAMPLE_RATE (descriptor->PortRangeHints[port_index].HintDescriptor)) {
    lower = descriptor->PortRangeHints[port_index].LowerBound * (LADSPA_Data) sample_rate;
    upper = descriptor->PortRangeHints[port_index].UpperBound * (LADSPA_Data) sample_rate;
  } else {
    lower = descriptor->PortRangeHints[port_index].LowerBound;
    upper = descriptor->PortRangeHints[port_index].UpperBound;
  }

  if (!LADSPA_IS_HINT_BOUNDED_BELOW (descriptor->PortRangeHints[port_index].HintDescriptor)) {
    lower = -100.0;
  }

  if (!LADSPA_IS_HINT_BOUNDED_ABOVE (descriptor->PortRangeHints[port_index].HintDescriptor)) {
    upper = 100.0;
  }
  
  if (!(lower < upper)) {
/*    printf("%s: lower !< upper! (%f, %f, %s, %s, %s)\n", __FUNCTION__,
           descriptor->PortRangeHints[port_index].LowerBound,
           descriptor->PortRangeHints[port_index].UpperBound,
           LADSPA_IS_HINT_SAMPLE_RATE (descriptor->PortRangeHints[port_index].HintDescriptor) ? "normal" : "sample rate",
           LADSPA_IS_HINT_BOUNDED_BELOW (descriptor->PortRangeHints[port_index].HintDescriptor) ? "use lower" : "do not use lower",
           LADSPA_IS_HINT_BOUNDED_ABOVE (descriptor->PortRangeHints[port_index].HintDescriptor) ? "use upper" : "do not use upper"
           );*/
    if (!LADSPA_IS_HINT_BOUNDED_ABOVE (descriptor->PortRangeHints[port_index].HintDescriptor)) {
      lower = upper - 100;
    } else {
      upper = lower + 100;
    }
  }
  widget = gtk_spin_button_new_with_range ((gdouble) lower, (gdouble) upper, 1.0);
  gtk_spin_button_set_numeric (GTK_SPIN_BUTTON (widget), TRUE);
  gtk_spin_button_set_digits (GTK_SPIN_BUTTON (widget), 0);
  
  return widget;
}

/** creates a string for the a control port label, possibly splitting
    from "foo (bar=baz,foobar=foobaz)" to "foo\nbar=baz\nfoobar=foobaz" */
const char * create_port_label (const char * port_name) {
  static GString * string = NULL;
  gchar * start;
  gchar * str;
  gchar * end;
  
  if (!string) {
    string = g_string_new ("");
  }
  
  str = g_strdup (port_name);
  
  start = strchr (str, '(');
  if (start) {
    end = strchr (start, '=');
    if (end) {
      *(start - 1) = '\0';
      g_string_printf (string, "%s", str);
      
      start++;
      do {
        end = strchr (start, ',');
        if (!end) {
          end = strchr (start, ')');
          if (end) {
            *end = '\0';
          }
          
          g_string_append_printf (string, "\n%s", start);
          g_free (str);
          return string->str;
        }
        
        *end = '\0';
        g_string_append_printf (string, "\n%s", start);
        
        end++;
        while (*end == ' ') end++;
        start = end;
      } while (1);
    }
  }

  g_string_printf (string, "%s", port_name);
  
  g_free (str);

  return string->str;
}
  
void create_mono_control_table_row (plugin_slot_t * plugin_slot,
                                    unsigned long port_index,
                                    port_controls_holder_t * port_controls_holder,
                                    lff_t ** control_fifos) {
  gint i;
  const LADSPA_Descriptor * descriptor;
  
  descriptor = plugin_slot->plugin->descriptor;

  /* lock control */  
  port_controls_holder->locked = TRUE;
  port_controls_holder->lock = gtk_toggle_button_new_with_label ("Lock");
  gtk_widget_show (port_controls_holder->lock);
  g_signal_connect (G_OBJECT (port_controls_holder->lock), "toggled",
                    G_CALLBACK (control_lock_callback), port_controls_holder);
  gtk_table_attach (GTK_TABLE (plugin_slot->control_table),
                    port_controls_holder->lock,
                    /* left-most column */   6, /* actually the 7th column; it's an index */
                    /* right-most column */  7,
                    /* upper-most row */     port_index,
                    /* lower-most row */     port_index+1,
                    /* x pack options */     0,
                    /* y pack options */     0,
                    /* x padding */          0,
                    /* y padding */          0);
  
  for (i = 0; i < 2; i++) {
    /* control fifo */
    port_controls_holder->controls.mono[i].control_fifo = control_fifos[i];
    
    /* label */
    port_controls_holder->controls.mono[i].label =
      gtk_label_new (create_port_label (descriptor->PortNames[port_index]));
    gtk_label_set_justify (GTK_LABEL(port_controls_holder->controls.mono[i].label),
                           GTK_JUSTIFY_CENTER);
    gtk_widget_show (port_controls_holder->controls.mono[i].label);
    gtk_table_attach (GTK_TABLE (plugin_slot->control_table),
                      port_controls_holder->controls.mono[i].label,
                      /* left-most column */   i * 3,
                      /* right-most column */  (i * 3)+1,
                      /* upper-most row */     port_index,
                      /* lower-most row */     port_index+1,
                      /* x pack options */     GTK_FILL,
                      /* y pack options */     0,
                      /* x padding */          3,
                      /* y padding */          0);
    
    /* create the controls */
    switch (port_controls_holder->type) {
      case JR_CTRL_FLOAT: 
        port_controls_holder->controls.mono[i].control =
          create_float_control (descriptor, port_index);
        g_signal_connect (G_OBJECT (port_controls_holder->controls.mono[i].control), "value-changed",
                          G_CALLBACK (control_float_callback), port_controls_holder);
        
        port_controls_holder->controls.mono[i].text =
          gtk_entry_new ();
        gtk_entry_set_max_length (GTK_ENTRY(port_controls_holder->controls.mono[i].text), TEXT_BOX_CHARS); 
        gtk_widget_set_size_request (port_controls_holder->controls.mono[i].text, TEXT_BOX_WIDTH, -1);
        gtk_widget_show (port_controls_holder->controls.mono[i].text);
        g_object_set_data (G_OBJECT (port_controls_holder->controls.mono[i].text),
                           "jack-rack-mono-channel", GINT_TO_POINTER(i));
        g_signal_connect (G_OBJECT (port_controls_holder->controls.mono[i].text), "activate",
                          G_CALLBACK (control_float_text_callback), port_controls_holder);
        break;
      
      case JR_CTRL_INT:
        port_controls_holder->controls.mono[i].control =
          create_int_control (descriptor, port_index);
        g_signal_connect (G_OBJECT (port_controls_holder->controls.mono[i].control), "value-changed",
                          G_CALLBACK (control_int_callback), port_controls_holder);
        port_controls_holder->controls.mono[i].text = NULL;
        break;
      
      case JR_CTRL_BOOL:
        port_controls_holder->controls.mono[i].control =
          gtk_toggle_button_new_with_label ("On");
        g_signal_connect (G_OBJECT (port_controls_holder->controls.mono[i].control), "toggled",
                          G_CALLBACK (control_bool_callback), port_controls_holder);
        port_controls_holder->controls.mono[i].text = NULL;
        break;
    }
    
    g_object_set_data (G_OBJECT (port_controls_holder->controls.mono[i].control),
                       "jack-rack-mono-channel", GINT_TO_POINTER(i));
    
    gtk_widget_show (port_controls_holder->controls.mono[i].control);
    
    /* pack the controls */
    switch (port_controls_holder->type) {
      case JR_CTRL_FLOAT:
        gtk_table_attach (GTK_TABLE (plugin_slot->control_table),
                          port_controls_holder->controls.mono[i].control,
                          /* left-most column */   (i * 3) + 1,
                          /* right-most column */  (i * 3) + 2,
                          /* upper-most row */     port_index,
                          /* lower-most row */     port_index+1,
                          /* x pack options */     GTK_EXPAND|GTK_FILL,
                          /* y pack options */     0,
                          /* x padding */          0,
                          /* y padding */          0);
        gtk_table_attach (GTK_TABLE (plugin_slot->control_table),
                          port_controls_holder->controls.mono[i].text,
                          /* left-most column */   (i * 3) + 2,
                          /* right-most column */  (i * 3) + 3, 
                          /* upper-most row */     port_index,
                          /* lower-most row */     port_index+1,
                          /* x pack options */     GTK_FILL,
                          /* y pack options */     0,
                          /* x padding */          0,
                          /* y padding */          0);
        break;
        
      case JR_CTRL_INT:
      case JR_CTRL_BOOL:
        gtk_table_attach (GTK_TABLE (plugin_slot->control_table),
                          port_controls_holder->controls.mono[i].control,
                          /* left-most column */   (i * 3) + 1,
                          /* right-most column */  (i * 3) + 3, 
                          /* upper-most row */     port_index,
                          /* lower-most row */     port_index+1,
                          /* x pack options */     GTK_EXPAND|GTK_FILL,
                          /* y pack options */     0,
                          /* x padding */          0,
                          /* y padding */          0);
      break;
    }
  }
  
}

void create_stereo_control_table_row (plugin_slot_t * plugin_slot,
                                      unsigned long port_index,
                                      port_controls_holder_t * port_controls_holder,
                                      lff_t * control_fifo) {
  const LADSPA_Descriptor * descriptor;
  
  descriptor = plugin_slot->plugin->descriptor;
  
  /* control fifo */
  port_controls_holder->controls.stereo.control_fifo = control_fifo;
  
  /* label */
  port_controls_holder->controls.stereo.label =
    gtk_label_new (create_port_label (descriptor->PortNames[port_index]));
  gtk_label_set_justify (GTK_LABEL(port_controls_holder->controls.stereo.label),
                         GTK_JUSTIFY_CENTER);
  gtk_widget_show (port_controls_holder->controls.stereo.label);
  gtk_table_attach (GTK_TABLE (plugin_slot->control_table),
                    port_controls_holder->controls.stereo.label,
                    /* left-most column */   0,
                    /* right-most column */  1,
                    /* upper-most row */     port_index,
                    /* lower-most row */     port_index+1,
                    /* x pack options */     GTK_FILL,
                    /* y pack options */     0,
                    /* x padding */          3,
                    /* y padding */          0);
  
  
  switch (port_controls_holder->type) {
    case JR_CTRL_FLOAT:
      port_controls_holder->controls.stereo.control =
        create_float_control (descriptor, port_index);
      g_signal_connect (G_OBJECT (port_controls_holder->controls.stereo.control), "value-changed",
                        G_CALLBACK (control_float_callback), port_controls_holder);
                        
      port_controls_holder->controls.stereo.text = gtk_entry_new ();
      gtk_entry_set_max_length (GTK_ENTRY(port_controls_holder->controls.stereo.text), TEXT_BOX_CHARS);
      gtk_widget_set_size_request (port_controls_holder->controls.stereo.text, TEXT_BOX_WIDTH, -1);
      gtk_widget_show (port_controls_holder->controls.stereo.text);
      g_signal_connect (G_OBJECT (port_controls_holder->controls.stereo.text), "activate",
                        G_CALLBACK (control_float_text_callback), port_controls_holder);
      break;
      
    case JR_CTRL_INT:
      port_controls_holder->controls.stereo.control =
        create_int_control (descriptor, port_index);
      g_signal_connect (G_OBJECT (port_controls_holder->controls.stereo.control), "value-changed",
                        G_CALLBACK (control_int_callback), port_controls_holder);
      port_controls_holder->controls.stereo.text = NULL;
      break;
      
    case JR_CTRL_BOOL:
      port_controls_holder->controls.stereo.control =
        gtk_toggle_button_new_with_label ("On");
      g_signal_connect (G_OBJECT (port_controls_holder->controls.stereo.control), "toggled",
                        G_CALLBACK (control_bool_callback), port_controls_holder);
      port_controls_holder->controls.stereo.text = NULL;
      break;
  }
  
  gtk_widget_show (port_controls_holder->controls.stereo.control);
    
  /* pack the controls */
  switch (port_controls_holder->type) {
    case JR_CTRL_FLOAT:
      gtk_table_attach (GTK_TABLE (plugin_slot->control_table),
                        port_controls_holder->controls.stereo.control,
                        /* left-most column */   1,
                        /* right-most column */  2, 
                        /* upper-most row */     port_index,
                        /* lower-most row */     port_index+1,
                        /* x pack options */     GTK_EXPAND|GTK_FILL,
                        /* y pack options */     0,
                        /* x padding */          0,
                        /* y padding */          0);
      gtk_table_attach (GTK_TABLE (plugin_slot->control_table),
                        port_controls_holder->controls.stereo.text,
                        /* left-most column */   2,
                        /* right-most column */  3, 
                        /* upper-most row */     port_index,
                        /* lower-most row */     port_index+1,
                        /* x pack options */     GTK_FILL,
                        /* y pack options */     0,
                        /* x padding */          0,
                        /* y padding */          0);
      break;
        
    case JR_CTRL_INT:
    case JR_CTRL_BOOL:
      gtk_table_attach (GTK_TABLE (plugin_slot->control_table),
                        port_controls_holder->controls.stereo.control,
                        /* left-most column */   1,
                        /* right-most column */  3, 
                        /* upper-most row */     port_index,
                        /* lower-most row */     port_index+1,
                        /* x pack options */     GTK_EXPAND|GTK_FILL,
                        /* y pack options */     0,
                        /* x padding */          0,
                        /* y padding */          0);
      break;
  }
}

void create_control_table (plugin_slot_t * plugin_slot,
                           ladspa_chain_link_t * chain_link) {
  plugin_t * plugin;
  unsigned long i;
  port_controls_holder_t * port_controls_holder;
  lff_t * control_fifos[2];
  
  plugin = plugin_slot->plugin;
  
  plugin_slot->control_table =
    gtk_table_new (plugin->control_port_count,
                   plugin->mono ? 7 : 3,
                   FALSE);
                   
  gtk_widget_show (plugin_slot->control_table);
  gtk_box_pack_start (GTK_BOX(plugin_slot->main_vbox),
                      plugin_slot->control_table, FALSE, FALSE, 0);
  
  /* make the controls array */
  plugin_slot->port_control_holders =
    xmalloc (sizeof (port_controls_holder_t) * plugin->control_port_count);
  
  /* contruct the control rows */
  for (i = 0; i < plugin->control_port_count; i++) {
    port_controls_holder = plugin_slot->port_control_holders + i;
    port_controls_holder->plugin_slot = plugin_slot;
    port_controls_holder->mono = plugin->mono;
    port_controls_holder->port_index = i;
    
    /* get the port control type from the hints */
    if (LADSPA_IS_HINT_TOGGLED (plugin_slot->plugin->descriptor->
        PortRangeHints[plugin->control_port_indicies[i]].HintDescriptor)) {
      port_controls_holder->type = JR_CTRL_BOOL;
    } else if (LADSPA_IS_HINT_INTEGER (plugin_slot->plugin->descriptor->
               PortRangeHints[plugin->control_port_indicies[i]].HintDescriptor)) {
      port_controls_holder->type = JR_CTRL_INT;
    } else {
      port_controls_holder->type = JR_CTRL_FLOAT;
    }
    
    if (plugin->mono) {
      control_fifos[0] = chain_link->holder.mono_pair.left->control_fifos + i;
      control_fifos[1] = chain_link->holder.mono_pair.right->control_fifos + i;
      create_mono_control_table_row (plugin_slot,
                                     plugin->control_port_indicies[i],
                                     port_controls_holder,
                                     control_fifos);
    } else {
      create_stereo_control_table_row (plugin_slot,
                                       plugin->control_port_indicies[i],
                                       port_controls_holder,
                                       chain_link->holder.stereo->control_fifos + i);
    }
  }
}

void set_controls_to_defaults (port_controls_holder_t * port_control_holders,
                               plugin_settings_t * settings) {
  unsigned long i, plugin_index;
  port_controls_holder_t * port_controls_holder = NULL;
  int j;
  port_controls_t * port_controls;
  
  plugin_index = port_control_holders->plugin_slot->plugin->collection_index;
  
  if (port_control_holders->plugin_slot->plugin->mono) {
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(port_control_holders->plugin_slot->lock_all),
                                    settings[plugin_index].lock_all ? TRUE : FALSE);
  }
  
  for (i = 0; i < port_control_holders->plugin_slot->plugin->control_port_count; i++) {
    port_controls_holder = port_control_holders + i;

    if (port_control_holders->plugin_slot->plugin->mono) {
      /* lock button */
      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(port_controls_holder->lock),
                                    settings[plugin_index].locks[i]);
      gtk_widget_set_sensitive (GTK_WIDGET(port_controls_holder->lock),
                                settings[plugin_index].lock_all ? FALSE : TRUE );
      
      if (settings[plugin_index].lock_all) {
        port_controls_holder->locked = TRUE;
      } else {
        port_controls_holder->locked = settings[plugin_index].locks[i];
      }
      
      port_controls = port_controls_holder->controls.mono;
    } else {
      port_controls = &port_controls_holder->controls.stereo;
    }
    
    for (j = 0; j < (port_controls_holder->plugin_slot->plugin->mono ? 2 : 1); j++) {
      switch (port_controls_holder->type) {
        case JR_CTRL_FLOAT: {
          gchar * str;
          gtk_range_set_value (GTK_RANGE (port_controls[j].control),
                               settings[plugin_index].control_values[j][i]);
          str = g_strdup_printf("%f", settings[plugin_index].control_values[j][i]);
          gtk_entry_set_text (GTK_ENTRY (port_controls[j].text), str);
          g_free (str);
          break;
        }
        case JR_CTRL_INT:
          gtk_spin_button_set_value (GTK_SPIN_BUTTON (port_controls[j].control),
                                     settings[plugin_index].control_values[j][i]);
          break;
        case JR_CTRL_BOOL:
          gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (port_controls[j].control),
                                        settings[plugin_index].control_values[j][i] > 0.0 ? TRUE : FALSE);
          break;
      }
    }
  }
}

void init_plugin_settings (plugin_settings_t * plugin_settings,
                           plugin_collection_t * collection) {
  unsigned long i = 0, j;
  plugin_t * plugin;
  
  for (plugin = collection->plugins;
       plugin;
       plugin = plugin->next, i++) {
     
    
    /* allocate arrays */
    plugin_settings[i].control_values[0] =
      xmalloc (sizeof (LADSPA_Data) * plugin->control_port_count);
      
    if (plugin->mono) {
      plugin_settings[i].control_values[1] =
        xmalloc (sizeof (LADSPA_Data) * plugin->control_port_count);
      plugin_settings[i].locks =
        xmalloc (sizeof (gboolean) * plugin->control_port_count);
      plugin_settings[i].lock_all = TRUE;
    } else {
      plugin_settings[i].control_values[1] = NULL;
      plugin_settings[i].locks = NULL;
    }
    
    /* set to defaults */
    if (plugin->mono) {
      for (j = 0; j < plugin->control_port_count; j++) {
        plugin_settings[i].control_values[0][j] =
        plugin_settings[i].control_values[1][j] =
          collection->default_plugin_settings[i][j];
        plugin_settings[i].locks[j] = TRUE;
      }
    } else {
      for (j = 0; j < plugin->control_port_count; j++) {
        plugin_settings[i].control_values[0][j] =
          collection->default_plugin_settings[i][j];
      }
    }
  }
}

void create_lock_all_button (plugin_slot_t * plugin_slot) {
  plugin_slot->lock_all = gtk_toggle_button_new_with_label ("Lock All");
  g_signal_connect (G_OBJECT (plugin_slot->lock_all), "toggled",
                    G_CALLBACK (slot_lock_all_callback), plugin_slot);
  gtk_widget_show (plugin_slot->lock_all);
  gtk_box_pack_end (GTK_BOX (plugin_slot->top_controls),
                    plugin_slot->lock_all, FALSE, FALSE, 0);
}
         

void init_plugin_slot (plugin_slot_t * plugin_slot,
                       plugin_collection_t * collection,
                       plugin_t * plugin,
                       ladspa_chain_link_t * chain_link) {
  gchar * str;

  plugin_slot->plugin = plugin;
  plugin_slot->enabled = FALSE;
  plugin_slot->next = NULL;
  plugin_slot->prev = NULL;
  
  /* create plugin settings array */
  plugin_slot->plugin_settings =
    xmalloc (sizeof (plugin_settings_t) * collection->plugin_count);
  init_plugin_settings (plugin_slot->plugin_settings, collection);
             
  /* create the gui */
  /* main vbox */
  plugin_slot->main_vbox = gtk_vbox_new (FALSE, 0);
  gtk_widget_show (plugin_slot->main_vbox);
  
  /* top control box */
  plugin_slot->top_controls = gtk_hbox_new (FALSE, 0);
  gtk_widget_show (plugin_slot->top_controls);
  gtk_box_pack_start (GTK_BOX (plugin_slot->main_vbox),
                      plugin_slot->top_controls, FALSE, FALSE, 0);
  
  /* plugin selector menu thingy */
  plugin_slot->plugin_selector = gtk_button_new_with_label (plugin->name);
  gtk_widget_show (plugin_slot->plugin_selector);
  plugin_slot->plugin_menu =
    create_plugin_menu (collection, G_CALLBACK (slot_change_callback), plugin_slot);
  g_signal_connect_swapped (G_OBJECT (plugin_slot->plugin_selector), "event",
	                    G_CALLBACK (plugin_button_callback),
	                    G_OBJECT (plugin_slot->plugin_menu));  
  gtk_box_pack_start (GTK_BOX (plugin_slot->top_controls),
                      plugin_slot->plugin_selector, FALSE, FALSE, 0);
  
  /* slot up */
  plugin_slot->slot_up = gtk_button_new ();
  gtk_widget_show (plugin_slot->slot_up);
  str = g_strconcat(PKGDATADIR, "/jr-up-arrow.png", NULL);
  plugin_slot->slot_up_image = gtk_image_new_from_file (str);
  g_free (str);
  g_object_set_data (G_OBJECT(plugin_slot->slot_up), "jack-rack-up", GINT_TO_POINTER(1));
  gtk_widget_show (plugin_slot->slot_up_image);
  gtk_container_add (GTK_CONTAINER (plugin_slot->slot_up), plugin_slot->slot_up_image);
  g_signal_connect (G_OBJECT (plugin_slot->slot_up), "clicked",
                      G_CALLBACK (slot_move_callback), plugin_slot);
  gtk_box_pack_start (GTK_BOX (plugin_slot->top_controls),
                      plugin_slot->slot_up, FALSE, FALSE, 0);
  
  /* slot remove */
  plugin_slot->slot_remove = gtk_button_new ();
  gtk_widget_show (plugin_slot->slot_remove);
  str = g_strconcat(PKGDATADIR, "/jr-cross.png", NULL);
  plugin_slot->slot_remove_image = gtk_image_new_from_file (str);
  g_free (str);
  gtk_widget_show (plugin_slot->slot_remove_image);
  gtk_container_add (GTK_CONTAINER (plugin_slot->slot_remove), plugin_slot->slot_remove_image);
  g_signal_connect (G_OBJECT (plugin_slot->slot_remove), "clicked",
                      G_CALLBACK (slot_remove_callback), plugin_slot);
  gtk_box_pack_start (GTK_BOX (plugin_slot->top_controls),
                      plugin_slot->slot_remove, FALSE, FALSE, 0);
  
  /* slot down */
  plugin_slot->slot_down = gtk_button_new ();
  gtk_widget_show (plugin_slot->slot_down);
  str = g_strconcat(PKGDATADIR, "/jr-down-arrow.png", NULL);
  plugin_slot->slot_down_image = gtk_image_new_from_file (str);
  g_free (str);
  g_object_set_data (G_OBJECT(plugin_slot->slot_down), "jack-rack-up", GINT_TO_POINTER(0));
  gtk_widget_show (plugin_slot->slot_down_image);
  gtk_container_add (GTK_CONTAINER (plugin_slot->slot_down), plugin_slot->slot_down_image);
  g_signal_connect (G_OBJECT (plugin_slot->slot_down), "clicked",
                      G_CALLBACK (slot_move_callback), plugin_slot);
  gtk_box_pack_start (GTK_BOX (plugin_slot->top_controls),
                      plugin_slot->slot_down, FALSE, FALSE, 0);
  
  /* enable button */
  plugin_slot->enable = gtk_toggle_button_new_with_label ("Enable");
  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (plugin_slot->enable), FALSE);
  g_signal_connect (G_OBJECT (plugin_slot->enable), "button-press-event",
                    G_CALLBACK (slot_ablise_callback), plugin_slot);
  gtk_widget_show (plugin_slot->enable);
  gtk_box_pack_start (GTK_BOX (plugin_slot->top_controls),
                      plugin_slot->enable, FALSE, FALSE, 0);
  
  
  
  /* sort out the port controls */
  if (plugin->control_port_count > 0) {
  
    /* lock all button */
    if (plugin->mono) {
      create_lock_all_button (plugin_slot);
    } else {
      plugin_slot->lock_all = NULL;
    }
    
    create_control_table (plugin_slot, chain_link);
    
    set_controls_to_defaults (plugin_slot->port_control_holders,
                              plugin_slot->plugin_settings);
  } else {
    plugin_slot->lock_all = NULL;
    plugin_slot->control_table = NULL;
    plugin_slot->port_control_holders = NULL;
  }

  /* final seperator bar */  
  plugin_slot->separator = gtk_hseparator_new ();
  gtk_widget_show (plugin_slot->separator);
  gtk_box_pack_start (GTK_BOX (plugin_slot->main_vbox),
                      plugin_slot->separator, FALSE, FALSE, 4);
}

void add_plugin_slot (guiinfo_t * gui_info,
                      ladspa_chain_link_t * chain_link,
                      plugin_t * plugin) {
  plugin_slot_t * plugin_slot;
  
  /* create the plugin_slot */  
  plugin_slot = xmalloc (sizeof (plugin_slot_t));
  
  init_plugin_slot (plugin_slot, gui_info->plugin_collection, plugin, chain_link);
  
  if (gui_info->slots) {
    plugin_slot->prev = gui_info->slots_end;
    plugin_slot->prev->next = plugin_slot;
    plugin_slot->next = NULL;
    gui_info->slots_end = plugin_slot;
    plugin_slot->index = plugin_slot->prev->index + 1;
  } else {
    plugin_slot->next = NULL;
    plugin_slot->prev = NULL;
    gui_info->slots = plugin_slot;
    gui_info->slots_end = plugin_slot;
    plugin_slot->index = 0;
  }
  gui_info->slot_count++;
  
  gtk_box_pack_start (GTK_BOX(gui_info->plugin_box),
                      plugin_slot->main_vbox, FALSE, FALSE, 0);

}

void remove_plugin_slot (guiinfo_t * gui_info,
                         plugin_slot_t * plugin_slot) {
  unsigned long i;
  
  /* adjust pointers */
  if (plugin_slot->prev) {
    plugin_slot->prev->next = plugin_slot->next;
  } else {
    gui_info->slots = plugin_slot->next;
  }
  
  /* adjust pointers and indicies */
  if (plugin_slot->next) {
    plugin_slot_t * ps;
    
    plugin_slot->next->prev = plugin_slot->prev;
    
    for (ps = plugin_slot->next;
         ps;
         ps = ps->next) {
      ps->index--;
    }
  } else {
    gui_info->slots_end = plugin_slot->prev;
  }
  gui_info->slot_count--;
  
  /* destroy all the widgets */
  gtk_widget_destroy (plugin_slot->main_vbox);
  
  /* free up the memory */
  free (plugin_slot->port_control_holders);
  for (i = 0; i < gui_info->plugin_collection->plugin_count; i++) {
    free (plugin_slot->plugin_settings[i].control_values[0]);
    if (gui_info->plugin_collection->plugin_array[i]->mono) {
      free (plugin_slot->plugin_settings[i].control_values[1]);
      free (plugin_slot->plugin_settings[i].locks);
    }
  }
  free (plugin_slot->plugin_settings);

  free (plugin_slot);
}

void move_plugin_slot (guiinfo_t * gui_info,
                       plugin_slot_t * plugin_slot,
                       gint up) {
  plugin_slot_t * tmpslot;
  if (up) {
    if (!plugin_slot->prev) return;
    
    plugin_slot->index--;
    plugin_slot->prev->index++;
    
    if (plugin_slot->next) {
      plugin_slot->next->prev = plugin_slot->prev;
    } else {
      gui_info->slots_end = plugin_slot->prev;
    }
    tmpslot = plugin_slot->prev->prev;
    if (tmpslot) {
      tmpslot->next = plugin_slot;
    } else {
      gui_info->slots = plugin_slot;
    }
    plugin_slot->prev->next = plugin_slot->next;
    plugin_slot->prev->prev = plugin_slot;
    plugin_slot->next = plugin_slot->prev;
    plugin_slot->prev = tmpslot;
    
  } else {
    if (!plugin_slot->next) return;
    
    plugin_slot->index++;
    plugin_slot->next->index--;
    
    if (plugin_slot->prev) {
      plugin_slot->prev->next = plugin_slot->next;
    } else {
      gui_info->slots = plugin_slot->next;
    }
    tmpslot = plugin_slot->next->next;
    if (tmpslot) {
      tmpslot->prev = plugin_slot;
    } else {
      gui_info->slots_end = plugin_slot;
    }
    plugin_slot->next->prev = plugin_slot->prev;
    plugin_slot->next->next = plugin_slot;
    plugin_slot->prev = plugin_slot->next;
    plugin_slot->next = tmpslot;
  }
  
  gtk_box_reorder_child (GTK_BOX(gui_info->plugin_box),
                         plugin_slot->main_vbox,
                         plugin_slot->index);
}

void change_plugin_slot (unsigned long plugin_slot_index,
                         ladspa_chain_link_t * chain_link,
                         guiinfo_t * gui_info) {
  unsigned long i;
  plugin_slot_t * plugin_slot;
/*  ctrlmsg_t ctrlmsg; */
  
  /* find our plugin slot */
  for (i = 0, plugin_slot = gui_info->slots;
       i < plugin_slot_index;
       i++, plugin_slot = plugin_slot->next) {
  }
  
  /* sort out the lock all button */
  if (chain_link->mono &&
      chain_link->control_port_count > 0 &&
      !plugin_slot->lock_all) {
    create_lock_all_button (plugin_slot);
  }
  if ((!chain_link->mono && plugin_slot->lock_all) ||
      (chain_link->mono &&
       chain_link->control_port_count == 0 &&
       plugin_slot->lock_all)) {
    gtk_widget_destroy (plugin_slot->lock_all);
    plugin_slot->lock_all = NULL;
  }
  
  
  /* kill all the control stuff */
  if (plugin_slot->control_table) {
    gtk_widget_destroy (plugin_slot->control_table);
    free (plugin_slot->port_control_holders);
  }
  
  /* get rid of this so we can add a new one afterward */
  gtk_widget_destroy (plugin_slot->separator);
  
  gtk_button_set_label (GTK_BUTTON(plugin_slot->plugin_selector),
                        chain_link->plugin->name);
  
  plugin_slot->plugin = chain_link->plugin;
  plugin_slot->chain_link = chain_link;
  plugin_slot->enabled = FALSE;
  
  /* create the new port controls */
  if (chain_link->plugin->control_port_count > 0) {
  
    create_control_table (plugin_slot, chain_link);
  
    set_controls_to_defaults (plugin_slot->port_control_holders,
                              plugin_slot->plugin_settings);
  } else {
    plugin_slot->control_table = NULL;
    plugin_slot->port_control_holders = NULL;
  }
  
  /* add a new separator */
  plugin_slot->separator = gtk_hseparator_new ();
  gtk_widget_show (plugin_slot->separator);
  gtk_box_pack_start (GTK_BOX (plugin_slot->main_vbox),
                      plugin_slot->separator, FALSE, FALSE, 4);
  
  /* make sure the enable thingy is set right in the process() */
/*  if (plugin_slot->enabled) {
    ctrlmsg.type = CTRLMSG_ABLE;
    ctrlmsg.number = plugin_slot->index;
    ctrlmsg.pointer = GINT_TO_POINTER (plugin_slot->enabled);
    ctrlmsg.second_pointer = plugin_slot;
    
    lff_write (gui_info->gui_to_process, &ctrlmsg);
  }*/
  
  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (plugin_slot->enable), FALSE);
}

void destroy_gui (guiinfo_t * gui_info) {
  plugin_slot_t * plugin_slot, * next_plugin_slot;

  /* remove all the plugin slots */
  for (plugin_slot = gui_info->slots_end;
       plugin_slot;
       plugin_slot = next_plugin_slot) {
    next_plugin_slot = plugin_slot->prev;
    
    remove_plugin_slot (gui_info, plugin_slot);
  }
  
  /* destroy the rest of the widgets */
  gtk_widget_destroy (gui_info->main_window);
}

