/*
 *   jack-ladspa-host
 *    
 *   Copyright (C) Robert Ham 2002 (node@users.sourceforge.net)
 *    
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <gtk/gtk.h>
#define _GNU_SOURCE
#include <getopt.h>
#undef _GNU_SOURCE
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <stdlib.h>
#include <glib.h>

#include "lock-free-fifo.h"
#include "control-message.h"
#include "gui.h"
#include "plugin.h"
#include "config.h"
#include "callbacks.h"
#ifdef HAVE_LADCCA
#include <ladcca/ladcca.h>
#endif

plugin_collection_t global_plugin_collection;
guiinfo_t global_gui_info;
process_info_t global_process_info;

#ifdef HAVE_LADCCA
cca_client_t * cca_client = NULL;
#endif

void print_help () {
  printf("%s version %s\n", PACKAGE_NAME, PACKAGE_VERSION);
  printf("Copyright (C) 2002 Robert Ham (node@users.sourceforge.net)\n");
  printf("\n");
  printf("This program comes with ABSOLUTELY NO WARRANTY.  You are licensed to use it\n");
  printf("under the terms of the GNU General Public License, version 2 or later.  See\n");
  printf("the COPYING file that came with this software for details.\n");
  printf("\n");
#ifdef HAVE_LADCCA
  printf("Compiled with JACK %s, LADCCA %s\n", JACK_VERSION, LADCCA_VERSION);
#else
  printf("Compiled with JACK %s\n", JACK_VERSION);
#endif
  printf("\n");
  printf(" -h, --help                  Display this help info\n");
  printf(" -p, --pid-name              Use the pid in the JACK client name\n");
  printf(" -s, --string-name <string>  Use <string> in the JACK client name\n");
#ifdef HAVE_JACK_SET_SERVER_DIR
  printf(" -D, --tmpdir <dir>          Tell JACK to use <dir> for its temporary files\n");
#endif
  printf("\n");
}

int main (int argc, char ** argv) {
  char * client_name = NULL;
  int opt;
  const char * options = "hps:D:";
  struct option long_options[] = {
    { "help", 0, NULL, 'h' },
    { "pid-name", 0, NULL, 'p' },
    { "string-name", 1, NULL, 's' },
#ifdef HAVE_JACK_SET_SERVER_DIR
    { "tmpdir", 1, NULL, 'D' },
#endif
    { 0, 0, 0, 0 }
  };
#ifdef HAVE_LADCCA
  cca_args_t * cca_args;
  
  cca_args = cca_extract_args (&argc, &argv);
#endif
  
  /* for printing out memory usage at the end */
/*  g_mem_set_vtable (glib_mem_profiler_table); */


  gtk_set_locale ();
  gtk_init (&argc, &argv);
  
  while ((opt = getopt_long (argc, argv, options, long_options, NULL)) != -1) {
    switch (opt) {
      case 'h':
        print_help ();
        exit (0);
        break;
      case 's':
        if (!client_name) client_name = g_strdup_printf ("jack_rack_%s", optarg);
        break;
      case 'p':
        if (!client_name) client_name = g_strdup_printf ("jack_rack_%d", getpid());
        break;
#ifdef HAVE_JACK_SET_SERVER_DIR
      case 'D':
        jack_set_server_dir (optarg);
        break;
#endif
      case ':':
        fprintf (stderr, "missing option parameter\n");
        print_help ();
        exit (1);
        break;
      case '?':
        print_help ();
        exit (1);
        break;
    }
  }
  
#ifdef HAVE_LADCCA
  cca_client = cca_init (cca_args, "JACK Rack", CCA_Use_Jack, CCA_PROTOCOL (1,0));
#endif
  
  if (!client_name) {
        client_name = g_strdup ("jack_rack");
  }

  init_jack (&global_process_info, client_name);
  g_free (client_name);
  
  
  init_plugin_collection (&global_plugin_collection);

  global_gui_info.gui_to_process = global_process_info.gui_to_process;
  global_gui_info.process_to_gui = global_process_info.process_to_gui;
  global_gui_info.process_info = &global_process_info;
  
  create_gui (&global_plugin_collection, &global_gui_info);
  
  jack_on_shutdown (global_process_info.jack_client,
                    jack_shutdown_callback,
                    &global_gui_info);
  
  jack_activate (global_process_info.jack_client);
  
  gtk_main ();

  jack_client_close (global_process_info.jack_client);
  
  destroy_gui (&global_gui_info);
  
  destroy_process_info (&global_process_info);
  
  destroy_plugin_collection (&global_plugin_collection);
  
/*  g_mem_profile (); */
  
  return 0;
}

