/*
 *   jack-ladspa-host
 *    
 *   Copyright (C) Robert Ham 2002 (node@users.sourceforge.net)
 *    
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef __JLH_PLUGIN_H__
#define __JLH_PLUGIN_H__

#include <ladspa.h>
#include <glib-2.0/glib.h>

#include "chain.h"

typedef struct object_file {
  struct object_file * next;

  char * filename;
  void * dl_handle;
  
  LADSPA_Descriptor_Function get_descriptor;
  
  unsigned long plugin_count;
} object_file_t;


typedef struct plugin {
  struct plugin * next;
  
  object_file_t * object_file;
  const char * name;
  unsigned long index;
  const LADSPA_Descriptor * descriptor;
  gboolean mono;
  
  unsigned long control_port_count;
  unsigned long * control_port_indicies;
  
  unsigned long port_indicies[4];
  
  /** for output ports */
  LADSPA_Data unused_memory;
  
  /* the plugin's index in the plugin collection */
  unsigned long collection_index;
} plugin_t;


typedef struct plugin_collection_t {
  object_file_t * object_files;
  unsigned long object_file_count;
  
  plugin_t * plugins;
  unsigned long plugin_count;
  plugin_t ** plugin_array;
  
  LADSPA_Data ** default_plugin_settings;
  
} plugin_collection_t;

extern plugin_collection_t global_plugin_collection;

extern void init_plugin_collection (plugin_collection_t *);
extern LADSPA_Data ** duplicate_default_plugin_settings (plugin_collection_t * collection);
extern unsigned long find_plugin (plugin_t *, plugin_collection_t *);
extern void destroy_plugin_collection (plugin_collection_t *);

#endif /* __JLH_PLUGIN_H__ */
